"""
MAL-Toolbox Attack Graph Attacker Class
"""

from dataclasses import dataclass
from typing import List, ForwardRef

@dataclass
class Attacker:
    id: str
    entry_points: List[ForwardRef("node.AttackGraphNode")]
    reached_attack_steps: List[ForwardRef("node.AttackGraphNode")]
    node: ForwardRef("node.AttackGraphNode")

    def to_dict(self):
        attacker_dict = {
            "id": self.id,
            "entry_points": [entry_point.id for entry_point in
                self.entry_points],
            "reached_attack_steps": [attack_step.id for attack_step in
                self.reached_attack_steps],
            "node": self.node.id
        }

        return attacker_dict



"""
MAL-Toolbox Language Classes Factory Module
"""

import python_jsonschema_objects as pjs
import json
import logging

logger = logging.getLogger(__name__)

class LanguageClassesFactory:
    def __init__(self, langspec):
        self.langspec = langspec
        self.json_schema = {}

    def generate_assets(self):
        """
        Generate JSON Schema for the assets in the language specification.
        """
        for asset in self.langspec['assets']:
            logger.debug(f'Creating {asset["name"]} asset JSON '\
            'schema entry.')
            asset_json_entry = {'title': asset['name'], 'type': 'object',
                'properties': {}}
            asset_json_entry['properties']['metaconcept'] = \
                {
                    'type' : 'string',
                    'default': asset['name']
                }
            if asset['superAsset']:
                asset_json_entry['allOf'] = [{
                    '$ref': '#/definitions/LanguageAsset/definitions/' + asset['superAsset']
                }]
            for defense in filter(lambda item: item['type'] == 'defense',
                asset['attackSteps']):
                if defense['ttc'] and defense['ttc']['name'] == 'Enabled':
                    default_defense_value = 1.0
                else:
                    default_defense_value = 0.0
                asset_json_entry['properties'][defense['name']] = \
                    {
                        'type' : 'number',
                        'minimum' : 0,
                        'maximum' : 1,
                        'default': default_defense_value
                    }
            self.json_schema['definitions']['LanguageAsset']['definitions']\
                [asset['name']] = asset_json_entry
            self.json_schema['definitions']['LanguageAsset']['oneOf'].\
                append({'$ref': '#/definitions/LanguageAsset/definitions/'
                + asset['name']})

    def generate_associations(self):
        """
        Generate JSON Schema for the associations in the language specification.
        """
        def create_association_entry(assoc):
            logger.debug(f'Creating {assoc["name"]} association JSON '\
            'schema entry.')
            assoc_json_entry = {'title': assoc['name'], 'type': 'object',
                'properties': {}}

            create_association_field(assoc, assoc_json_entry, 'left')
            create_association_field(assoc, assoc_json_entry, 'right')
            return assoc_json_entry

        def create_association_with_subentries(assoc):
            if (assoc['name'] not in self.json_schema['definitions']\
                ['LanguageAssociation']['definitions']):
                logger.info('Multiple associations with the same '\
                    f'name, {assoc["name"]}, exist. '\
                    'Creating subentries for each one.')
                self.json_schema['definitions']['LanguageAssociation']\
                ['definitions'][assoc['name']] =\
                    {
                        'title': assoc['name'],
                        'type': 'object',
                        'oneOf': [],
                        'definitions': {}
                    }
                self.json_schema['definitions']['LanguageAssociation']['oneOf'].\
                    append({'$ref': '#/definitions/LanguageAssociation/definitions/'
                    + assoc['name']})

            assoc_json_subentry = create_association_entry(assoc)
            subentry_name = assoc['name'] + '_' + assoc['leftAsset'] + '_' \
                + assoc['rightAsset']

            logger.info(f'Creating {subentry_name} subentry association.')
            assoc_json_subentry['title'] = subentry_name
            self.json_schema['definitions']['LanguageAssociation']\
                ['definitions'][assoc['name']]['definitions'][subentry_name] = assoc_json_subentry
            self.json_schema['definitions']['LanguageAssociation']\
                ['definitions'][assoc['name']]['oneOf'].append(
                    {'$ref': '#/definitions/LanguageAssociation/definitions/' \
                    + assoc['name'] + '/definitions/' + subentry_name})

        def create_association_field(assoc, assoc_json_entry, position):
            assoc_json_entry['properties'][assoc[position + 'Field']] = \
                {
                    'type' : 'array',
                    'items' :
                        {
                        '$ref':
                        '#/definitions/LanguageAsset/definitions/' +
                            assoc[position + 'Asset']
                        }
                }
            if assoc[position + 'Multiplicity']['max']:
                assoc_json_entry['properties'][assoc[position + 'Field']]\
                    ['maxItems'] = assoc[position + 'Multiplicity']['max']


        for assoc in self.langspec['associations']:
            count = len(list(filter(lambda temp_assoc: temp_assoc['name'] ==
                assoc['name'], self.langspec['associations'])))
            if count > 1:
                # If there are multiple associations with the same name we
                # will need to create separate entries for each using their
                # fieldnames.
                assoc_json_entry = create_association_with_subentries(assoc)
            else:
                assoc_json_entry = create_association_entry(assoc)
                self.json_schema['definitions']['LanguageAssociation']\
                    ['definitions'][assoc['name']] = assoc_json_entry
                self.json_schema['definitions']['LanguageAssociation']['oneOf'].\
                    append({'$ref': '#/definitions/LanguageAssociation/' +
                    'definitions/' + assoc['name']})

    def create_classes(self):
        """
        Create classes based on the language specification.
        """

        # First, we have to translate the language specification into a JSON
        # schema. Initialize the overall JSON schema structure.
        self.json_schema = {
            'title': 'LanguageObject',
            'type': 'object',
            'oneOf':[
                {'$ref': '#/definitions/LanguageAsset'},
                {'$ref': '#/definitions/LanguageAssociation'}
            ],
            'definitions': {}}
        self.json_schema['definitions']['LanguageAsset'] = {
            'title': 'LanguageAsset',
            'type': 'object',
            'oneOf': [],
            'definitions': {}}
        self.json_schema['definitions']['LanguageAssociation'] = {
            'title': 'LanguageAssociation',
            'type': 'object',
            'oneOf': [],
            'definitions': {}}

        self.generate_assets()
        self.generate_associations()
        logger.debug(json.dumps(self.json_schema, indent = 2))

        # Once we have the JSON schema we create the actual classes.
        builder = pjs.ObjectBuilder(self.json_schema)
        self.ns = builder.build_classes(standardize_names=False)



"""
MAL-Toolbox Main Module
"""

import sys

import logging
import json

import maltoolbox
import maltoolbox.cl_parser
from maltoolbox.language import classes_factory
from maltoolbox.language import specification
from maltoolbox.model import model
from maltoolbox.attackgraph import attackgraph
from maltoolbox.attackgraph.analyzers import apriori
from maltoolbox.ingestors import neo4j

logger = logging.getLogger(__name__)

def main():
    """Main routine of maltoolbox command line interface."""
    args = maltoolbox.cl_parser.parse_args(sys.argv[1:])
    cmd_params = vars(args)
    logger.info('Received the following command line parameters:\n' +
        json.dumps(cmd_params, indent = 2))

    match(cmd_params['command']):
        case 'gen_ag':
            model_filename = cmd_params['model']
            langspec_filename = cmd_params['language']

            # Load language specification and generate classes based on it.
            lang_spec = specification. \
                load_language_specification_from_mar(langspec_filename)
            if maltoolbox.log_configs['langspec_file']:
                specification.save_language_specification_to_json(lang_spec,
                    maltoolbox.log_configs['langspec_file'])
            lang_classes_factory = \
                classes_factory.LanguageClassesFactory(lang_spec)
            lang_classes_factory.create_classes()

            # Load instance model.
            instance_model = model.Model('Test model', lang_spec,
                lang_classes_factory)
            instance_model.load_from_file(model_filename)
            if maltoolbox.log_configs['model_file']:
                instance_model.save_to_file( \
                    maltoolbox.log_configs['model_file'])

            graph = attackgraph.AttackGraph()
            result = graph.generate_graph(lang_spec, instance_model)
            if result > 0:
                logger.error('Attack graph generation failed!')
                print('Attack graph generation failed!')
                exit(result)

            apriori.calculate_viability_and_necessity(graph)

            graph.attach_attackers(instance_model)

            if maltoolbox.log_configs['attackgraph_file']:
                graph.save_to_file(
                    maltoolbox.log_configs['attackgraph_file'])

            if cmd_params['neo4j']:
                # Injest instance model and attack graph into Neo4J.
                logger.debug('Ingest model graph into Neo4J database.')
                neo4j.ingest_model(instance_model,
                    maltoolbox.neo4j_configs['uri'],
                    maltoolbox.neo4j_configs['username'],
                    maltoolbox.neo4j_configs['password'],
                    maltoolbox.neo4j_configs['dbname'],
                    delete=True)
                logger.debug('Ingest attack graph into Neo4J database.')
                neo4j.ingest_attack_graph(graph,
                    maltoolbox.neo4j_configs['uri'],
                    maltoolbox.neo4j_configs['username'],
                    maltoolbox.neo4j_configs['password'],
                    maltoolbox.neo4j_configs['dbname'],
                    delete=False)

        case _:
            logger.error(f'Received unknown command: {cmd_params["command"]}.')



"""
MAL-Toolbox Model Module
"""

import json
import logging

from dataclasses import dataclass
from typing import List

logger = logging.getLogger(__name__)

@dataclass
class Attacker:
    id: str = None
    name: str = None
    entry_points: List[tuple] = None

class Model:
    latestId = 0

    def __init__(self, name, lang_spec, lang_classes_factory):
        self.name = name
        self.assets = []
        self.associations = []
        self.attackers = []
        self.lang_spec = lang_spec
        self.lang_classes_factory = lang_classes_factory

    def add_asset(self, asset, asset_id: int = None):
        """
        Add an asset to the model.
        """
        if asset_id:
            asset.id = asset_id
        else:
            asset.id = self.latestId
        self.latestId = max(asset.id + 1, self.latestId)

        asset.associations = []
        if not hasattr(asset, 'name'):
            asset.name = asset.metaconcept + ':' + str(asset.id)
        self.assets.append(asset)

    def add_association(self, association):
        """
        Add an association to the model.
        """
        for prop in range(0, 2):
            for asset in getattr(association,
                list(vars(association)['_properties'])[prop]):
                assocs = list(asset.associations)
                assocs.append(association)
                asset.associations = assocs
        self.associations.append(association)

    def add_attacker(self, attacker, attacker_id: int = None):
        """
        Add an attacker to the model.
        """
        if attacker_id:
            attacker.id = attacker_id
        else:
            attacker.id = self.latestId
        self.latestId = max(attacker.id + 1, self.latestId)

        if not hasattr(attacker, 'name') or not attacker.name:
            attacker.name = 'Attacker:' + str(attacker.id)
        self.attackers.append(attacker)

    def get_asset_by_id(self, asset_id):
        """
        Find an asset in the model based on its id.

        Arguments:
        asset_id        - the id of the asset we are looking for

        Return:
        An asset matching the id if it exists in the model.
        """
        return next((asset for asset in self.assets if asset.id == asset_id),
            None)

    def get_attacker_by_id(self, attacker_id):
        """
        Find an attacker in the model based on its id.

        Arguments:
        attacker_id     - the id of the attacker we are looking for

        Return:
        An attacker matching the id if it exists in the model.
        """
        return next((attacker for attacker in self.attackers \
            if attacker.id == attacker_id), None)

    def get_associated_assets_by_field_name(self, asset, field_name):
        """
        Get a list of associated assets for an asset given a fieldname.

        Arguments:
        asset           - the asset whose fields we are interested in
        fieldname       - the field name we are looking for

        Return:
        A list of assets associated with the asset given that match the
        fieldname.
        """
        logger.debug(f'Get associated assets for asset '
            f'{asset.name}(id:{asset.id}) by field name {field_name}.')
        associated_assets = []
        for association in asset.associations:
            # Determine which two of the end points leads away from the asset.
            # This is particularly relevant for associations between two
            # assets of the same type.
            if asset in getattr(association,
                list(vars(association)['_properties'])[0]):
                elementName = list(vars(association)['_properties'])[1]
            else:
                elementName = list(vars(association)['_properties'])[0]

            if elementName == field_name:
                associated_assets.extend(getattr(association, elementName))

        return associated_assets

    def asset_to_json(self, asset):
        """
        Convert an asset to its JSON format.
        """
        defenses = {}
        logger.debug(f'Translating {asset.name} to json.')
        for defense in list(vars(asset)['_properties'])[1:]:
            defenseValue = getattr(asset, defense)
            logger.debug(f'Translating {defense}: {defenseValue} defense '\
                'to json.')
            if defenseValue != getattr(asset, defense).default():
            # Save the default values that are not the default ones.
                defenses[str(defense)] = str(defenseValue)
        return (str(asset.id), {
                'name': str(asset.name),
                'metaconcept': str(asset.metaconcept),
                'eid': str(asset.id),
                'defenses': defenses
                }
            )

    def association_to_json(self, association):
        """
        Convert an association to its JSON format.
        """
        firstElementName = list(vars(association)['_properties'])[0]
        secondElementName = list(vars(association)['_properties'])[1]
        firstElements = getattr(association, firstElementName)
        secondElements = getattr(association, secondElementName)
        json_association = {
            'metaconcept': type(association).__name__,
            'association': {
                str(firstElementName): [str(asset.id) for asset in firstElements],
                str(secondElementName): [str(asset.id) for asset in secondElements]
            }
        }
        return json_association

    def attacker_to_json(self, attacker):
        """
        Convert an attacker to its JSON format.
        """
        logger.debug(f'Translating {attacker.name} to json.')
        json_attacker = {
            'name': str(attacker.name),
            'entry_points': {},
        }
        for (asset, attack_steps) in attacker.entry_points:
            json_attacker['entry_points'][str(asset.id)] = {
                'attack_steps' : attack_steps
            }
        return (str(attacker.id), json_attacker)

    def model_to_json(self):
        """
        Convert the model to its JSON format.
        """
        logger.debug(f'Converting model to JSON format.')
        contents = {
            'metadata': {},
            'assets': {},
            'associations': [],
            'attackers' : {}
        }
        contents['metadata'] = {
            'name': self.name,
            'langVersion': self.lang_spec['defines']['version'],
            'langID': self.lang_spec['defines']['id'],
            'malVersion': '0.1.0-SNAPSHOT',
            'info': 'Created by the mal-toolbox model python module.'
        }

        logger.debug('Translating assets to json.')
        for asset in self.assets:
            (asset_id, asset_json) = self.asset_to_json(asset)
            contents['assets'][asset_id] = asset_json

        logger.debug('Translating associations to json.')
        for association in self.associations:
            assoc_json = self.association_to_json(association)
            contents['associations'].append(assoc_json)

        logger.debug('Translating attackers to json.')
        for attacker in self.attackers:
            (attacker_id, attacker_json) = self.attacker_to_json(attacker)
            contents['attackers'][attacker_id] = attacker_json
        return contents

    def save_to_file(self, filename):
        """
        Save model to a json file.

        Arguments:
        filename        - the name of the output file
        """

        logger.info(f'Saving model to {filename} file.')
        contents = self.model_to_json()
        fp = open(filename, 'w')
        json.dump(contents, fp, indent = 2)

    def load_from_file(self, filename):
        """
        Load model from a json file.

        Arguments:
        filename        - the name of the input file
        """
        logger.info(f'Loading model from {filename} file.')
        with open(filename, 'r', encoding='utf-8') as model_file:
            json_model = json.loads(model_file.read())

        self.name = json_model['metadata']['name']

        # Reconstruct the assets
        for asset_id in json_model['assets']:
            asset_object = json_model['assets'][asset_id]
            logger.debug(f'Loading asset from {filename}:\n' \
                + json.dumps(asset_object, indent = 2))
            asset = getattr(self.lang_classes_factory.ns,
                asset_object['metaconcept'])(name = asset_object['name'])
            for defense in asset_object['defenses']:
                setattr(asset, defense,
                    float(asset_object['defenses'][defense]))
            self.add_asset(asset, asset_id = int(asset_id))

        # Reconstruct the associations
        if 'associations' in json_model:
            for association_object in json_model['associations']:
                association = getattr(self.lang_classes_factory.ns,
                    association_object['metaconcept'])()
                for field in association_object['association']:
                    asset_list = []
                    for asset_id in association_object['association'][field]:
                        asset_list.append(self.get_asset_by_id(int(asset_id)))
                    setattr(association, field, asset_list)
                self.add_association(association)

        # Reconstruct the attackers
        if 'attackers' in json_model:
            attackers_info = json_model['attackers']
            for attacker_id in attackers_info:
                attacker = Attacker(name = attackers_info[attacker_id]['name'])
                attacker.entry_points = []
                for asset_id in attackers_info[attacker_id]['entry_points']:
                    attacker.entry_points.append(
                        (self.get_asset_by_id(int(asset_id)),
                        attackers_info[attacker_id]['entry_points']\
                            [asset_id]['attack_steps']))
                self.add_attacker(attacker, attacker_id = int(attacker_id))



"""
MAL-Toolbox Attack Graph Node Dataclass
"""

from dataclasses import dataclass
from typing import Any, List, Optional, ForwardRef

from maltoolbox.attackgraph import attacker

@dataclass
class AttackGraphNode:
    id: str = None
    type: str = None
    asset: Any = None
    name: str = None
    ttc: dict = None
    children: List[ForwardRef('AttackGraphNode')] = None
    parents: List[ForwardRef('AttackGraphNode')] = None
    defense_status: dict = None
    existence_status: bool = None
    is_viable: bool = True
    is_necessary: bool = True
    compromised_by: List[ForwardRef('Attacker')] = None
    extra: dict = None
    mitre_info: str = None
    tags: List[str] = None
    observations: dict = None

    def to_dict(self):
        node_dict = {
            'id': self.id,
            'type': self.type,
            'name': self.name,
            'ttc': self.ttc,
            'children': [child.id for child in self.children],
            'parents': [parent.id for parent in self.parents],
            'compromised_by': ['Attacker:' + attacker.id for attacker in \
                self.compromised_by]
        }

        if self.asset != None:
            node_dict['asset'] = self.asset.metaconcept + ':' \
                + str(self.asset.id)
        if self.defense_status != None:
            node_dict['defense_status'] = str(self.defense_status)
        if self.existence_status != None:
            node_dict['existence_status'] = str(self.existence_status)
        if self.is_viable != None:
            node_dict['is_viable'] = str(self.is_viable)
        if self.is_necessary != None:
            node_dict['is_necessary'] = str(self.is_necessary)
        if self.mitre_info != None:
            node_dict['mitre_info'] = str(self.mitre_info)
        if self.tags:
            node_dict['tags'] = str(self.tags)
        if self.observations != None:
            node_dict['observations'] = self.observations

        return node_dict



"""
MAL-Toolbox securiCAD Translator Module
"""

import zipfile
import json
import logging
import xml.etree.ElementTree as ET

from maltoolbox.model import model
from maltoolbox.language import specification

logger = logging.getLogger(__name__)

def load_model_from_scad_archive(scad_archive: str,
    lang_spec,
    lang_classes_factory) -> model.Model:
    """
    Reads a '.sCAD' archive generated by securiCAD representing an instance
    model and loads the information into a maltoobox.model.Model object.

    Arguments:
    scad_archive            - the path to a '.sCAD' archive
    lang_spec               - a dictionary containing the MAL language
                              specification
    lang_classes_factory    - a language classes factory that contains
                              the same classes defined by the
                              language specification

    Return:
    A maltoobox.model.Model object containing the instance model.
    """
    with zipfile.ZipFile(scad_archive, 'r') as archive:
        filelist = archive.namelist()
        model_file = next(filter(lambda x: ( x[-4:] == '.eom'), filelist))
        scad_model = archive.read(model_file)
        root = ET.fromstring(scad_model)

    instance_model = model.Model(scad_archive,
        lang_spec,
        lang_classes_factory)

    for child in root.iter('objects'):
        logger.debug(f'Loading asset from \"{scad_archive}\":\n' \
            + json.dumps(child.attrib, indent = 2))
        if child.attrib['metaConcept'] == 'Attacker':
            attacker_id = int(child.attrib['id'])
            attacker = model.Attacker()
            attacker.entry_points = []
            instance_model.add_attacker(attacker, attacker_id = attacker_id)
            continue

        if not hasattr(lang_classes_factory.ns,
            child.attrib['metaConcept']):
            logger.error(f'Failed to find {child.attrib["metaConcept"]} '
                'asset in language specification!')
            return None
        asset = getattr(lang_classes_factory.ns,
            child.attrib['metaConcept'])(name = child.attrib['name'])
        asset_id = int(child.attrib['id'])
        for subchild in child.iter('evidenceAttributes'):
            defense_name = subchild.attrib['metaConcept']
            defense_name = defense_name[0].lower() + defense_name[1:]
            for distrib in subchild.iter('evidenceDistribution'):
                distrib_type = distrib.attrib['type']
                for d in distrib.iter('parameters'):
                    if 'value' in d.attrib:
                        dist_value = d.attrib['value']
                        setattr(asset, defense_name, float(dist_value))
        instance_model.add_asset(asset, asset_id)

    for child in root.iter('associations'):
        logger.debug(f'Load association '
            f'(\"{child.attrib["sourceObject"]}\",'
            f'\"{child.attrib["targetObject"]}",'
            f'\"{child.attrib["targetProperty"]}\",'
            f'\"{child.attrib["sourceProperty"]}\") '
            f'from \"{scad_archive}\"')

        # Note: This is not a bug in the code. The fields and assets are
        # listed incorrectly in the securiCAD format where the source asset
        # matches the target field and vice versa.
        left_id = int(child.attrib['targetObject'])
        right_id = int(child.attrib['sourceObject'])
        attacker_id = None
        if child.attrib['sourceProperty'] == 'firstSteps':
            attacker_id = right_id
            target_id = left_id
            target_prop = child.attrib['targetProperty']
        elif child.attrib['targetProperty'] == 'firstSteps':
            attacker_id = left_id
            target_id = right_id
            target_prop = child.attrib['sourceProperty']

        if attacker_id:
            attacker = instance_model.get_attacker_by_id(attacker_id)
            if not attacker:
                logger.error(f'Failed to find attacker with id {attacker_id} '
                    'in model!')
                return None
            target_asset = instance_model.get_asset_by_id(target_id)
            if not target_asset:
                logger.error(f'Failed to find asset with id {target_id} '
                    'in model!')
                return None
            attacker.entry_points.append((target_asset,
                [target_prop.split('.')[0]]))
            continue

        left_asset = instance_model.get_asset_by_id(left_id)
        if not left_asset:
            logger.error(f'Failed to find asset with id {left_id} '
                'in model!')
            return None
        right_asset = instance_model.get_asset_by_id(right_id)
        if not right_asset:
            logger.error(f'Failed to find asset with id {right_id} '
                'in model!')
            return None

        # Note: This is not a bug in the code. The fields and assets are
        # listed incorrectly in the securiCAD format where the source asset
        # matches the target field and vice versa.
        left_field = child.attrib['sourceProperty']
        right_field = child.attrib['targetProperty']
        assoc_name = specification.get_association_by_fields_and_assets(
            lang_spec,
            left_field,
            right_field,
            left_asset.metaconcept,
            right_asset.metaconcept)
        logger.debug(f'Found \"{assoc_name}\" association.')

        if not assoc_name:
            logger.error(f'Failed to find '
                f'(\"{left_asset.metaconcept}\",'
                f'\"{right_asset.metaconcept}",'
                f'\"{left_field}\",'
                f'\"{right_field}\") '
                'association in language specification!')
            return None

        if not hasattr(lang_classes_factory.ns,
            assoc_name):
            logger.error(f'Failed to find {assoc_name} '
                'association in language specification!')
            return None

        assoc = getattr(lang_classes_factory.ns, assoc_name)()
        setattr(assoc, left_field, [left_asset])
        setattr(assoc, right_field, [right_asset])
        instance_model.add_association(assoc)

    return instance_model



# -*- encoding: utf-8 -*-
# MAL Toolbox v0.0.1
# Copyright 2023, Andrei Buhaiu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""
MAL-Toolbox Framework
"""

__title__ = 'maltoolbox'
__version__ = '0.0.19'
__authors__ = ['Andrei Buhaiu']
__license__ = 'Apache 2.0'
__docformat__ = 'restructuredtext en'

__all__ = ()

import os
import sys
import configparser
import logging

from pkg_resources import Requirement, resource_filename

ERROR_INCORRECT_CONFIG = 1

CONFIGFILE = resource_filename(Requirement.parse("mal-toolbox"),
    "maltoolbox/default.conf")

config = configparser.ConfigParser()
config.read(CONFIGFILE)

if 'logging' not in config:
    print('Config file is missing essential information, cannot proceed.')
    sys.exit(ERROR_INCORRECT_CONFIG)

for term in ['output_dir', 'log_file']:
    if term not in config['logging']:
        logger.critical('Config file is missing essential '\
            'information, cannot proceed.')
        print('Config file is missing essential information, cannot '\
            'proceed.')
        sys.exit(ERROR_INCORRECT_CONFIG)

log_configs = {
    'output_dir': config['logging']['output_dir'],
    'log_file': config['logging']['log_file'],
    'attackgraph_file': config['logging']['attackgraph_file'],
    'model_file': config['logging']['model_file'],
    'langspec_file': config['logging']['langspec_file'],
}

os.makedirs(log_configs['output_dir'], exist_ok = True)
logging.basicConfig(level=logging.DEBUG,
            format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
            datefmt='%m-%d %H:%M',
            filename=log_configs["log_file"],
            filemode='w')
logging.getLogger('python_jsonschema_objects').setLevel(logging.WARNING)

logger = logging.getLogger(__name__)

if 'neo4j' in config:
    for term in ['uri', 'username', 'password', 'dbname']:
        if term not in config['neo4j']:
            logger.critical('Config file is missing essential '\
                f'Neo4J information: {term}, cannot proceed.')
            print('Config file is missing essential '\
                f'Neo4J information: {term}, cannot proceed.')
            sys.exit(ERROR_INCORRECT_CONFIG)

    neo4j_configs = {
        'uri': config['neo4j']['uri'],
        'username': config['neo4j']['username'],
        'password': config['neo4j']['password'],
        'dbname': config['neo4j']['dbname'],
    }




"""
MAL-Toolbox Attack Graph Query Submodule

This submodule contains functions that analyze the information present in the
attack graph, but do not alter the structure or nodes in any way.
"""

import logging

from maltoolbox.attackgraph import attackgraph
from maltoolbox.attackgraph import attacker

logger = logging.getLogger(__name__)

def is_node_traversable_by_attacker(node, attacker) -> bool:
    """
    Return True or False depending if the node specified is traversable
    for the attacker given.

    Arguments:
    node        - the node we wish to evalute
    attacker    - the attacker whose traversability we are interested in
    """

    logger.debug(f'Evaluate if {node.id}, of type \'{node.type}\', is '\
        f'traversable by Attacker {attacker.id}')
    if not node.is_viable:
        return False

    match(node.type):
        case 'or':
            return True

        case 'and':
            for parent in node.parents:
                if parent not in attacker.reached_attack_steps and \
                     parent.is_necessary:
                    # If the parent is not present in the attacks steps
                    # already reached and is necessary.
                    return False
            return True

        case 'exist' | 'notExist' |  'defense':
            return False

        case _:
            logger.error(f'Unknown node type {node.type}.')
            return False

def get_attack_surface(graph: attackgraph.AttackGraph,
    attacker: attacker.Attacker):
    """
    Get the current attack surface of an attacker. This includes all of the
    viable children nodes of already reached attack steps that are of 'or'
    type and the 'and' type children nodes which have all of their necessary
    parents in the attack steps reached.

    Arguments:
    graph       - the attack graph
    attacker    - the Attacker whose attack surface is sought
    """
    logger.debug(f'Get the attack surface for Attacker {attacker.id}.')
    attack_surface = []
    for attack_step in attacker.reached_attack_steps:
        logger.debug('Determine attack surface stemming from '
            f'{attack_step.id} for Attacker {attacker.id}.')
        for child in attack_step.children:
            if is_node_traversable_by_attacker(child, attacker):
                attack_surface.append(child)
    return attack_surface

def get_defense_surface(graph: attackgraph.AttackGraph):
    """
    Get the defense surface. All non-suppressed defense steps that are not
    already fully enabled.

    Arguments:
    graph       - the attack graph
    """
    logger.debug(f'Get the defense surface.')
    return (node for node in graph.nodes if node.type == 'defense' and \
        'suppress' not in node.tags and \
        node.defense_status != 1.0)

def get_enabled_defenses(graph: attackgraph.AttackGraph):
    """
    Get the defenses already enabled. All non-suppressed defense steps that
    are already fully enabled.

    Arguments:
    graph       - the attack graph
    """
    logger.debug(f'Get the enabled defenses.')
    return (node for node in graph.nodes if node.type == 'defense' and \
        'suppress' not in node.tags and \
        node.defense_status == 1.0)




"""
MAL-Toolbox Attack Graph Module
"""

import logging
import json

from typing import List

from maltoolbox.language import specification
from maltoolbox.model import model
from maltoolbox.attackgraph import node
from maltoolbox.attackgraph import attacker

logger = logging.getLogger(__name__)

class AttackGraph:
    def __init__(self):
        self.nodes = []
        self.attackers = []


    def save_to_file(self, filename: str):
        """
        Save the attack graph to a json file.

        Arguments:
        filename        - the name of the output file
        """

        logger.info(f'Saving attack graph to {filename} file.')
        serialized_graph = []
        for ag_node in self.nodes:
            serialized_graph.append(ag_node.to_dict())
        with open(filename, 'w', encoding='utf-8') as file:
            json.dump(serialized_graph, file, indent=4)


    def load_from_file(self, filename: str, model: model.Model = None):
        """
        Load the attack graph model from a json file.

        Arguments:
        filename        - the name of the input file to parse
        model           - (optional) the instance model that the attack graph was
                          generated from. If this given then the attack graph node
                          and instance model asset link can be re-established. If
                          this argument is not given the attack graph will still
                          be created it will just omit the links to the assets.
        """

        logger.info(f'Loading attack graph from {filename} file.')
        if model:
            logger.info(f'Model(\'{model.name}\') was provided will attempt '
            'to establish links to assets.')
        else:
            logger.info('No model was provided therefore asset links will '
            'not be established.')

        with open(filename, 'r', encoding='utf-8') as file:
            serialized_graph = json.load(file)
        # Create all of the nodes in the imported attack graph.
        for node_dict in serialized_graph:
            ag_node = node.AttackGraphNode()
            ag_node.id = node_dict['id']
            ag_node.type = node_dict['type']
            ag_node.name = node_dict['name']
            ag_node.ttc = node_dict['ttc']
            ag_node.asset = None
            ag_node.children = []
            ag_node.parents = []
            ag_node.compromised_by = []

            ag_node.defense_status = str(node_dict['defense_status']) if \
                'defense_status' in node_dict else None
            ag_node.existence_status = str(node_dict['existence_status']) if \
                'existence_status' in node_dict else None
            ag_node.is_viable = str(node_dict['is_viable']) if \
                'is_viable' in node_dict else True
            ag_node.is_necessary = str(node_dict['is_necessary']) if \
                'is_necessary' in node_dict else True
            ag_node.mitre_info = str(node_dict['mitre_info']) if \
                'mitre_info' in node_dict else None
            if ag_node.name == 'firstSteps':
                # This is an attacker entry point node, recreate the attacker.
                attacker_id = ag_node.id.split(':')[1]
                ag_attacker = attacker.Attacker(
                    id = str(attacker_id),
                    entry_points = [],
                    reached_attack_steps = [],
                    node = ag_node
                )
                self.attackers.append(ag_attacker)
                ag_node.attacker = ag_attacker

            self.nodes.append(ag_node)

        # Re-establish links between nodes.
        for node_dict in serialized_graph:
            ag_node = self.get_node_by_id(node_dict['id'])
            if ag_node == None:
                logger.error(f'Failed to find node with id {node_dict["id"]}'
                    f' when loading from attack graph from file {filename}')

            for child_id in node_dict['children']:
                child = self.get_node_by_id(child_id)
                if child == None:
                    logger.error(f'Failed to find child node with id {child_id}'
                        f' when loading from attack graph from file {filename}')
                    return None
                ag_node.children.append(child)

                if hasattr(ag_node, 'attacker'):
                    # Relink the attacker related connections since the node
                    # is an attacker entry point node.
                    ag_attacker = ag_node.attacker
                    ag_attacker.entry_points.append(child)
                    ag_attacker.reached_attack_steps.append(child)
                    child.compromised_by.append(ag_attacker)

            for parent_id in node_dict['parents']:
                parent = self.get_node_by_id(parent_id)
                if parent == None:
                    logger.error('Failed to find parent node with id '
                        f'{parent_id} when loading from attack graph from '
                        f'file {filename}')
                    return None
                ag_node.parents.append(parent)

            # Also recreate asset links if model is available.
            if model and 'asset' in node_dict:
                asset = model.get_asset_by_id(
                    int(node_dict['asset'].split(':')[1]))
                if asset == None:
                    logger.error('Failed to find asset with id '
                        f'{node_dict["asset"]} when loading from attack graph '
                        f'from file {filename}')
                    return None
                ag_node.asset = asset
                if hasattr(asset, 'attack_step_nodes'):
                    attack_step_nodes = list(asset.attack_step_nodes)
                    attack_step_nodes.append(ag_node)
                    asset.attack_step_nodes = attack_step_nodes
                else:
                    asset.attack_step_nodes = [ag_node]


    def get_node_by_id(self, node_id: str) -> node.AttackGraphNode:
        """
        Return the attack node that matches the id provided.

        Arguments:
        node_id     - the id of the attack graph none we are looking for

        Return:
        The attack step node that matches the given id.
        """

        logger.debug(f'Looking up node with id {node_id}')
        return next((ag_node for ag_node in self.nodes \
            if ag_node.id == node_id), None)


    def attach_attackers(self, model: model.Model):
        """
        Create attackers and their entry point nodes and attach them to the
        relevant attack step nodes and to the attackers.

        Arguments:
        model       - the instance model containing the attackers
        """

        logger.info(f'Attach attackers from \'{model.name}\' model to the '
            'graph.')
        for attacker_info in model.attackers:
            attacker_node = node.AttackGraphNode(
                    id = 'Attacker:' + str(attacker_info.id) + ':firstSteps',
                    type = 'or',
                    asset = None,
                    name = 'firstSteps',
                    ttc = None,
                    children = [],
                    parents = [],
                    compromised_by = []
            )

            ag_attacker = attacker.Attacker(
                id = str(attacker_info.id),
                entry_points = [],
                reached_attack_steps = [],
                node = attacker_node
            )
            attacker_node.attacker = ag_attacker
            self.attackers.append(ag_attacker)

            for (asset, attack_steps) in attacker_info.entry_points:
                for attack_step in attack_steps:
                    attack_step_id = asset.metaconcept + ':' \
                        + str(asset.id) + ':' + attack_step
                    ag_node = self.get_node_by_id(attack_step_id)
                    if not ag_node:
                        logger.warning('Failed to find attacker entry point '
                            + attack_step_id + ' for Attacker:'
                            + ag_attacker.id + '.')
                        continue
                    ag_node.compromised_by.append(ag_attacker)
                    ag_attacker.entry_points.append(ag_node)
                    logger.debug(f'Attach Attacker \'{ag_attacker.id}\' to '
                        f'{ag_node.id}.')

            ag_attacker.reached_attack_steps = ag_attacker.entry_points
            attacker_node.children = ag_attacker.entry_points
            self.nodes.append(attacker_node)


    def process_step_expression(self, lang: dict, model: model.Model,
        target_assets: List, step_expression: dict):
        """
        Recursively process an attack step expression.

        Arguments:
        lang            - a dictionary representing the MAL language specification
        model           - a maltoolbox.model.Model instance from which the attack
                          graph was generated
        target_assets   - the list of assets that this step expression should apply
                          to. Initially it will contain the asset to which the
                          attack step belongs
        step_expression - a dictionary containing the step expression

        Return:
        A tuple pair containing a list of all of the target assets and the name of
        the attack step.
        """
        logger.debug('Processing Step Expression:\n' \
            + json.dumps(step_expression, indent = 2))

        match (step_expression['type']):
            case 'attackStep':
                # The attack step expression just adds the name of the attack
                # step. All other step expressions only modify the target assets.
                return (target_assets, step_expression['name'])

            case 'union' | 'intersection' | 'difference':
                # The set operators are used to combine the left hand and right
                # hand targets accordingly.
                lh_targets, lh_attack_steps = self.process_step_expression(
                    lang, model, target_assets, step_expression['lhs'])
                rh_targets, rh_attack_steps = self.process_step_expression(
                    lang, model, target_assets, step_expression['rhs'])

                new_target_assets = []
                match (step_expression['type']):
                    case 'union':
                        new_target_assets = lh_targets
                        for ag_node in rh_targets:
                            if next((lnode for lnode in new_target_assets \
                                if lnode.id != ag_node.id), None):
                                new_target_assets.append(ag_node)

                    case 'intersection':
                        for ag_node in rh_targets:
                            if next((lnode for lnode in new_target_assets \
                                if lnode.id == ag_node.id), None):
                                new_target_assets.append(ag_node)

                    case 'difference':
                        new_target_assets = lh_targets
                        for ag_node in lh_targets:
                            if next((rnode for rnode in rh_target_assets \
                                if rnode.id != ag_node.id), None):
                                new_target_assets.remove(ag_node)

                return (new_target_assets, None)

            case 'variable':
                # Fetch the step expression associated with the variable from
                # the language specification and resolve that.
                for target_asset in target_assets:
                    if (hasattr(target_asset, 'metaconcept')):
                        variable_step_expr = specification.\
                            get_variable_for_class_by_name(lang,
                            target_asset.metaconcept, step_expression['name'])
                        return self.process_step_expression(
                            lang, model, target_assets, variable_step_expr)

                    else:
                        logger.error('Requested variable from non-asset'
                            f'target node: {target_asset} which cannot be'
                            'resolved.')
                return ([], None)

            case 'field':
                # Change the target assets from the current ones to the associated
                # assets given the specified field name.
                new_target_assets = []
                for target_asset in target_assets:
                    new_target_assets.extend(model.\
                        get_associated_assets_by_field_name(target_asset,
                            step_expression['name']))
                return (new_target_assets, None)

            case 'transitive':
                # The transitive expression is very similar to the field
                # expression, but it proceeds recursively until no target is
                # found and it and it sets the new targets to the entire list
                # of assets identified during the entire transitive recursion.
                new_target_assets = []
                for target_asset in target_assets:
                    new_target_assets.extend(model.\
                        get_associated_assets_by_field_name(target_asset,
                            step_expression['stepExpression']['name']))
                if new_target_assets:
                    (additional_assets, _) = self.process_step_expression(
                        lang, model, new_target_assets, step_expression)
                    new_target_assets.extend(additional_assets)
                    return (new_target_assets, None)
                else:
                    return ([], None)

            case 'subType':
                new_target_assets = []
                for target_asset in target_assets:
                    (assets, _) = self.process_step_expression(
                        lang, model, target_assets, step_expression['stepExpression'])
                    new_target_assets.extend(assets)

                selected_new_target_assets = (asset for asset in \
                    new_target_assets if specification.extends_asset(
                        lang,
                        asset.metaconcept,
                        step_expression['subType']))
                return (selected_new_target_assets, None)

            case 'collect':
                # Apply the right hand step expression to left hand step
                # expression target assets.
                lh_targets, _ = self.process_step_expression(
                    lang, model, target_assets, step_expression['lhs'])
                return self.process_step_expression(lang, model, lh_targets,
                    step_expression['rhs'])


            case _:
                logger.error('Unknown attack step type: '
                    f'{step_expression["type"]}')
                return ([], None)


    def generate_graph(self, lang: dict, model: model.Model):
        """
        Generate attack graph starting from a model instance
        and a MAL language specification

        Arguments:
        model           - a maltoolbox.model.Model instance
        lang            - a dictionary representing the MAL language specification
        """

        # First, generate all of the nodes of the attack graph.
        for asset in model.assets:
            logger.debug(f'Generating attack steps for asset {asset.name} which '\
                f'is of class {asset.metaconcept}.')
            attack_step_nodes = []
            attack_steps = specification.get_attacks_for_class(lang,
                asset.metaconcept)
            for attack_step_name, attack_step_attribs in attack_steps.items():
                logger.debug('Generating attack step node for '\
                    f'{attack_step_name}.')

                defense_status = None
                existence_status = None
                node_id = asset.metaconcept + ':' + str(asset.id) + ':' + attack_step_name

                match (attack_step_attribs['type']):
                    case 'defense':
                        # Set the defense status for defenses
                        defense_status = getattr(asset, attack_step_name)
                        logger.debug('Setting the defense status of '\
                            f'{node_id} to {defense_status}.')

                    case 'exist' | 'notExist':
                        # Resolve step expression associated with (non-)existence
                        # attack steps.
                        (target_assets, attack_step) = self.process_step_expression(
                            lang,
                            model,
                            [asset],
                            attack_step_attribs['requires']['stepExpressions'][0])
                        # If the step expression resolution yielded the target
                        # assets then the required assets exist in the model.
                        existence_status = target_assets != []

                mitre_info = attack_step_attribs['meta']['mitre'] if 'mitre' in\
                    attack_step_attribs['meta'] else None
                ag_node = node.AttackGraphNode(
                    id = node_id,
                    type = attack_step_attribs['type'],
                    asset = asset,
                    name = attack_step_name,
                    ttc = attack_step_attribs['ttc'],
                    children = [],
                    parents = [],
                    defense_status = defense_status,
                    existence_status = existence_status,
                    is_viable = True,
                    is_necessary = True,
                    mitre_info = mitre_info,
                    tags = attack_step_attribs['tags'],
                    compromised_by = []
                )
                ag_node.attributes = attack_step_attribs
                attack_step_nodes.append(ag_node)
                self.nodes.append(ag_node)
            asset.attack_step_nodes = attack_step_nodes

        # Then, link all of the nodes according to their associations.
        for ag_node in self.nodes:
            logger.debug('Determining children for attack step '\
                f'{ag_node.id}.')
            step_expressions = \
                ag_node.attributes['reaches']['stepExpressions'] if \
                    ag_node.attributes['reaches'] else []

            for step_expression in step_expressions:
                # Resolve each of the attack step expressions listed for this
                # attack step to determine children.
                (target_assets, attack_step) = self.process_step_expression(lang,
                     model, [ag_node.asset], step_expression)
                for target in target_assets:
                    target_node_id = target.metaconcept + ':' \
                        + str(target.id) + ':' + attack_step
                    target_node = self.get_node_by_id(target_node_id)
                    if not target_node:
                        logger.error('Failed to find targed node ' \
                        f'{target_node_id} to link with for attack step ' \
                        f'{ag_node.id}!')
                        print('Failed to find targed node ' \
                        f'{target_node_id} to link with for attack step ' \
                        f'{ag_node.id}!')
                        return 1
                    ag_node.children.append(target_node)
                    target_node.parents.append(ag_node)

        return 0



"""
MAL-Toolbox Language Specification Module

"""

import logging
import json
import zipfile
import copy

logger = logging.getLogger(__name__)

def load_language_specification_from_mar(mar_archive: str) -> dict:
    """
    Read a ".mar" archive provided by malc (https://github.com/mal-lang/malc)
    and return a dictionary representing a MAL language structure

    Arguments:
    mar_archive     -   the path to a ".mar" archive

    Return:
    A dictionary representing the language specification
    """

    logger.info(f'Load language specfication from \'{mar_archive}\' mar archive.')
    with zipfile.ZipFile(mar_archive, 'r') as archive:
        langspec = archive.read('langspec.json')
        return json.loads(langspec)

def load_language_specification_from_json(json_file: str) -> dict:
    """
    Read a MAL language JSON specification file

    Arguments:
    file_spec       - a language specification file that can be for example
                      provided by malc (https://github.com/mal-lang/malc)

    Return:
    A dictionary representing the language specification
    """

    logger.info(f'Load language specfication from \'{json_file}\'.')
    with open(json_file, 'r', encoding='utf-8') as spec:
        data = spec.read()
    return json.loads(data)


def save_language_specification_to_json(lang_spec: dict, filename: str) -> dict:
    """
    Save a MAL language specification dictionary to a JSON file

    Arguments:
    lang_spec       - a dictionary containing the MAL language specification
    filename        - the JSON filename where the language specification will
                      be written
    """

    logger.info(f'Save language specfication to {filename}.')

    with open(filename, 'w', encoding='utf-8') as file:
        json.dump(lang_spec, file, indent=4)


def get_attacks_for_class(lang_spec: dict, asset_type: str) -> dict:
    """
    Get all Attack Steps for a specific Class

    Arguments:
    lang_spec       - a dictionary containing the MAL language specification
    asset_type      - a string representing the class for which we want to list
                      the possible attack steps

    Return:
    A dictionary representing the set of possible attacks for the specified
    class. Each key in the dictionary is an attack name and is associated
    with a dictionary containing other characteristics of the attack such as
    type of attack, TTC distribution, child attack steps and other information
    """
    attacks = {}
    asset = next((asset for asset in lang_spec['assets'] if asset['name'] == \
        asset_type), None)
    if not asset:
        logger.error(f'Failed to find asset type {asset_type} when '\
            'looking for attack steps.')
        return None

    logger.debug(f'Get attack steps for {asset["name"]} asset from '\
        'language specification.')
    if asset['superAsset']:
        logger.debug(f'Asset extends another one, fetch the superclass '\
            'attack steps for it.')
        attacks = get_attacks_for_class(lang_spec, asset['superAsset'])

    for attack in asset['attackSteps']:
        if attack['name'] not in attacks:
            attacks[attack['name']] = copy.deepcopy(attack)
        else:
            if not attack['reaches']:
                # This attack step does not lead to any attack steps
                continue
            if attack['reaches']['overrides'] == True:
                attacks[attack['name']] = copy.deepcopy(attack)
            else:
                attacks[attack['name']]['reaches']['stepExpressions'].\
                    extend(attack['reaches']['stepExpressions'])

    return attacks

def get_associations_for_class(lang_spec: dict, asset_type: str) -> dict:
    """
    Get all Associations for a specific Class

    Arguments:
    lang_spec       - a dictionary containing the MAL language specification
    asset_type      - a string representing the class for which we want to list
                      the associations

    Return:
    A dictionary representing the set of associations for the specified
    class. Each key in the dictionary is an attack name and is associated
    with a dictionary containing other characteristics of the attack such as
    type of attack, TTC distribution, child attack steps and other information
    """
    logger.debug(f'Get associations for {asset_type} asset from '\
        'language specification.')
    associations = []

    asset = next((asset for asset in lang_spec['assets'] if asset['name'] == \
        asset_type), None)
    if not asset:
        logger.error(f'Failed to find asset type {asset_type} when '\
            'looking for associations.')
        return None

    if asset['superAsset']:
        logger.debug(f'Asset extends another one, fetch the superclass '\
            'associations for it.')
        associations.extend(get_associations_for_class(lang_spec,
            asset['superAsset']))
    assoc_iter = (assoc for assoc in lang_spec['associations'] \
        if assoc['leftAsset'] == asset_type or \
            assoc['rightAsset'] == asset_type)
    assoc = next(assoc_iter, None)
    while (assoc):
        associations.append(assoc)
        assoc = next(assoc_iter, None)

    return associations

def get_association_by_fields_and_assets(lang_spec: dict,
    first_field: str,
    second_field: str,
    first_asset: str,
    second_asset: str) -> str:
    """
    Get an association based on its field names and asset types

    Arguments:
    lang_spec       - a dictionary containing the MAL language specification
    first_field     - a string containing the first field
    second_field    - a string containing the second field
    first_asset     - a string representing the first asset type
    second_asset    - a string representing the second asset type

    Return:
    The name of the association matching the fieldnames and asset types.
    None if there is no match.
    """
    for assoc in lang_spec['associations']:
        logger.debug(f'Compare (\"{first_asset}\",'
            f'\"{first_field}\",'
            f'\"{second_asset}\",'
            f'\"{second_field}\") to '
            f'(\"{assoc["leftAsset"]}\",'
            f'\"{assoc["leftField"]}\",'
            f'\"{assoc["rightAsset"]}\",'
            f'\"{assoc["rightField"]}\").')

        # If the asset and fields match either way we accept it as a match.
        if assoc['leftField'] == first_field and \
            assoc['rightField'] == second_field and \
            extends_asset(lang_spec, first_asset, assoc['leftAsset']) and \
            extends_asset(lang_spec, second_asset, assoc['rightAsset']):
            return assoc['name']
        if assoc['leftField'] == second_field and \
            assoc['rightField'] == first_field and \
            extends_asset(lang_spec, second_asset, assoc['leftAsset']) and \
            extends_asset(lang_spec, first_asset, assoc['rightAsset']):
            return assoc['name']

    return None

def get_variable_for_class_by_name(lang_spec: dict, asset_type: str,
    variable_name:str) -> dict:
    """
    Get a variables for a specific asset type by name.
    NOTE: Variables are the ones specified in MAL through `let` statements

    Arguments:
    lang_spec       - a dictionary containing the MAL language specification
    asset_type      - a string representing the type of asset which contains
                    the variable
    variable_name   - the name of the variable to search for

    Return:
    A dictionary representing the step expressions for the specified variable.
    """

    asset = next((asset for asset in lang_spec['assets'] if asset['name'] == \
        asset_type), None)
    if not asset:
        logger.error(f'Failed to find asset type {asset_type} when '\
            'looking for variable.')
        return None

    variable_dict = next((variable for variable in \
        asset['variables'] if variable['name'] == variable_name), None)
    if not variable_dict:
        if asset['superAsset']:
            variable_dict = get_variable_for_class_by_name(lang_spec,
                asset['superAsset'], variable_name)
        if variable_dict:
            return variable_dict
        else:
            logger.error(f'Failed to find variable {variable_name} in '\
                f'{asset_type}\'s language specification.')
        return None

    return variable_dict['stepExpression']

def extends_asset(lang_spec: dict, asset, target_asset):
    """
    Check if an asset extends the target asset through inheritance.

    Arguments:
    lang_spec       - a dictionary containing the MAL language specification
    asset           - the asset name we wish to evaluate
    target_asset    - the target asset name we wish to evaluate if it
                      is extended

    Return:
    True if this asset extends the target_asset via inheritance.
    False otherwise.
    """

    logger.debug(f'Check if {asset} extends {target_asset} via inheritance.')

    if asset == target_asset:
        return True

    asset_dict = next((asset_info for asset_info in lang_spec['assets'] \
        if asset_info['name'] == asset), None)
    if not asset_dict:
        logger.error(f'Failed to find asset type {asset} when '\
            'looking for variable.')
        return False
    if asset_dict['superAsset']:
        if asset_dict['superAsset'] == target_asset:
            return True
        else:
            return extends_asset(lang_spec, asset_dict['superAsset'],
                target_asset)

    return False




"""
MAL-Toolbox Command Line Parsing Module
"""

import sys
import argparse
import logging
from typing import Sequence

from maltoolbox import __version__

logger = logging.getLogger(__name__)

def parse_args(args: Sequence[str]) -> argparse.Namespace:
    """
    This function parses the arguments which have been passed from the command
    line. It returns an argparse parser object.

    Arguments:
    args - the list of arguments passed from the command line in the sys.argv
           format

    Return:
    A parser with the provided arguments, which can be used in a simpler format
    """
    parser = argparse.ArgumentParser(
        prog='maltoolbox',
        description='Generate Attack Graphs from MAL specifications')

    parser.add_argument(
        '--version',
        action='version',
        version=f'%(prog)s {__version__}')

    subparsers = parser.add_subparsers(help='commands', dest='command')
    subparsers.required = True

    gen_ag_parser = subparsers.add_parser(
        'gen_ag',
        help='Generate an attack graph from an instance model from a json ' \
            'file and language specification from a mar archive.')
    help_parser = subparsers.add_parser(
        'help', help='Show help for a particular command')

    gen_ag_parser.add_argument(
        'model',
        help='Path to the instance model json file',
        type=str,
    )

    gen_ag_parser.add_argument(
        'language',
        help='Path to the language specification ".mar" archive',
        type=str,
    )

    gen_ag_parser.add_argument(
        '--neo4j',
        help='Injest attack graph and instance model into a local Neo4j ' \
            'instance',
        action='store_true'
    )

    help_parser.add_argument(
        'cmd',
        help='Name of command to get help for',
        nargs='?')

    if len(args) == 0:
        parser.print_help(sys.stderr)
        logger.error('Received no arugments will print help and exit.')
        sys.exit(1)

    parsed_args = parser.parse_args()
    if parsed_args.command == 'help':
        if not parsed_args.cmd:
            parser.print_help(sys.stderr)
        else:
            try:
                subparsers.choices[parsed_args.cmd].print_help()
            except KeyError:
                logger.error(f'Unknown command name {parsed_args.cmd}')
                print(f'Unknown command name {parsed_args.cmd}')
                print(
                    f'Valid commands are: {", ".join(subparsers.choices.keys())}')

        sys.exit(1)

    return parsed_args



"""
MAL-Toolbox Attack Graph Apriori Analyzer Submodule

This submodule contains analyzers that are relevant before attackers are even
connected to the attack graph.
Currently these are:
- Viability = Determine if a node can be traversed under any circumstances or
  if the model structure makes it unviable.
- Necessity = Determine if a node is necessary for the attacker or if the
  model structure means it is not needed(it behaves as if it were already
  compromised) to compromise children attack steps.
"""

import logging

from maltoolbox.attackgraph import attackgraph
from maltoolbox.attackgraph import attacker
from maltoolbox.attackgraph import node

logger = logging.getLogger(__name__)

def propagate_viability_from_node(node: node.AttackGraphNode):
    """
    Arguments:
    node        - the attack graph node from which to propagate the viable
                  status
    """
    logger.debug(f'Propagate viability from {node.id} with viability'
        f' status {node.is_viable}.')
    for child in node.children:
        original_value = child.is_viable
        if child.type == 'or':
            child.is_viable = False
            for parent in child.parents:
                child.is_viable = child.is_viable or parent.is_viable
        if child.type == 'and':
            child.is_viable = False

        if child.is_viable != original_value:
            propagate_viability_from_node(child)

def propagate_necessity_from_node(node: node.AttackGraphNode):
    """
    Arguments:
    node        - the attack graph node from which to propagate the necessary
                  status
    """
    logger.debug(f'Propagate necessity from {node.id} with necessity'
        f' status {node.is_necessary}.')
    for child in node.children:
        original_value = child.is_necessary
        if child.type == 'or':
            child.is_necessary = False
        if child.type == 'and':
            child.is_necessary = False
            for parent in child.parents:
                child.is_necessary = child.is_necessary or parent.is_necessary

        # TODO: Update TTC for child attack step before if it is not necessary
        # before propagating it further.
        if child.is_necessary != original_value:
            propagate_necessity_from_node(child)

def calculate_viability_and_necessity(graph: attackgraph.AttackGraph):
    """
    Arguments:
    node        - the attack graph for which we wish to determine the
                  viability and necessity statuses for the nodes.
    """
    for node in graph.nodes:
        match (node.type):
            case 'exist':
                node.is_viable = node.existence_status
                node.is_necessary = not node.existence_status
            case 'notExist':
                node.is_viable = not node.existence_status
                node.is_necessary = node.existence_status
            case 'defense':
                node.is_viable = node.defense_status != 1.0
                node.is_necessary = node.defense_status != 0.0
            case _:
                pass

        if not node.is_viable:
            propagate_viability_from_node(node)
        if not node.is_necessary:
            propagate_necessity_from_node(node)



"""
MAL-Toolbox Neo4j Ingestor Module
"""

import logging

from py2neo import Graph, Node, Relationship, Subgraph

logger = logging.getLogger(__name__)

def ingest_attack_graph(graph,
        uri,
        username,
        password,
        dbname,
        delete=False
) -> None:
    """
    Ingest an attack graph into a neo4j database

    Arguments:
    graph                - the attackgraph provided by the atkgraph.py module.
    uri                  - the URI to a running neo4j instance
    username             - the username to login on Neo4J
    password             - the password to login on Neo4J
    dbname               - the selected database
    delete               - if True, the previous content of the database is deleted
                           before ingesting the new attack graph
    """


    g = Graph(uri=uri, user=username, password=password, name=dbname)
    if delete:
        g.delete_all()

    nodes = {}
    rels = []
    for node in graph.nodes:
        node_dict = node.to_dict()
        nodes[node.id] = Node(node_dict['asset'] if 'asset' in node_dict else
            node_dict['id'],
            name=node_dict['name'],
            full_name=node_dict['id'],
            type=node_dict['type'],
            ttc=str(node_dict['ttc']),
            is_necessary = str(node.is_necessary),
            is_viable = str(node.is_viable),
            compromised_by = str(node_dict['compromised_by']))


    for node in graph.nodes:
        for child in node.children:
            rels.append(Relationship(nodes[node.id], nodes[child.id]))

    subgraph = Subgraph(list(nodes.values()), rels)

    tx = g.begin()
    tx.create(subgraph)
    g.commit(tx)


def ingest_model(model,
        uri,
        username,
        password,
        dbname,
        delete=False
) -> None:
    """
    Ingest an instance model graph into a Neo4J database

    Arguments:
    model                - the instance model dictionary as provided by the model.py module
    uri                  - the URI to a running neo4j instance
    username             - the username to login on Neo4J
    password             - the password to login on Neo4J
    dbname               - the selected database
    delete               - if True, the previous content of the database is deleted
                           before ingesting the new attack graph
    """
    g = Graph(uri=uri, user=username, password=password, name=dbname)
    if delete:
        g.delete_all()

    nodes = {}
    rels = []

    for asset in model.assets:
        nodeid = asset.name

        nodes[str(asset.id)] = Node(str(asset.metaconcept),
                name=str(asset.name),
                asset_id=str(asset.id),
                metaconcept=str(asset.metaconcept))

    for assoc in model.associations:
        firstElementName = list(vars(assoc)['_properties'])[0]
        secondElementName = list(vars(assoc)['_properties'])[1]
        firstElements = getattr(assoc, firstElementName)
        secondElements = getattr(assoc, secondElementName)
        for first_asset in firstElements:
            for second_asset in secondElements:
                rels.append(Relationship(nodes[str(first_asset.id)],
                    str(firstElementName),
                    nodes[str(second_asset.id)]))
                rels.append(Relationship(nodes[str(second_asset.id)],
                    str(secondElementName),
                    nodes[str(first_asset.id)]))


    subgraph = Subgraph(list(nodes.values()), rels)

    tx = g.begin()
    tx.create(subgraph)
    g.commit(tx)




"""
MAL-Toolbox Language Graph Module
"""

import logging
import json

from dataclasses import dataclass
from typing import Any, List, Optional, ForwardRef

from maltoolbox.language import specification


logger = logging.getLogger(__name__)

@dataclass
class LanguageGraphAsset:
    name: str = None
    associations: List[ForwardRef('LanguageGraphAssociation')] = None
    attack_steps: List[ForwardRef('LanguageGraphAttackStep')] = None
    description: dict = None
    # MAL languages currently do not support multiple inheritance, but this is
    # futureproofing at its most hopeful.
    super_assets: list = None
    sub_assets: list = None

    def to_dict(self):
        node_dict = {
            'name': self.name,
            'associations': [],
            'attack_steps': [],
            'description': self.description,
            'super_assets': [],
            'sub_assets': []
        }

        for assoc in self.associations:
            node_dict['associations'].append((assoc.name,
                assoc.fields[0][1],
                assoc.fields[1][1]))
        for attack_step in self.attack_steps:
            node_dict['attack_steps'].append(attack_step.name)
        for super_asset in self.super_assets:
            node_dict['super_assets'].append(super_asset.name)
        for sub_asset in self.sub_assets:
            node_dict['sub_assets'].append(sub_asset.name)
        return node_dict

    def subclasses(self, target_asset):
        """
        Check if an asset extends the target asset through inheritance.

        Arguments:
        target_asset    - the target asset we wish to evaluate if this asset
                          extends

        Return:
        True if this asset extends the target_asset via inheritance.
        False otherwise.
        """
        current_assets = [self]
        while (current_assets):
            current_asset = current_assets.pop()
            if current_asset == target_asset:
                return True
            current_assets.extend(current_asset.super_assets)
        return False

@dataclass
class LanguageGraphAssociation:
    name: str = None
    # The left and right field of the association are tuples that contain a
    # reference to the asset type, the field name, minimum, and maximum
    # cardinality. Together the two form a tuple pair.
    fields: tuple[tuple[ForwardRef('LanguageGraphAsset'), str, int, int],
        tuple[ForwardRef('LanguageGraphAsset'), str, int, int]] = ()
    description: dict = None

    def to_dict(self):
        node_dict = {
            'name': self.name,
            'left': {
                'asset': self.fields[0][0].name,
                'fieldname': self.fields[0][1],
                'min': self.fields[0][2],
                'max': self.fields[0][3]
            },
            'right': {
                'asset': self.fields[1][0].name,
                'fieldname': self.fields[1][1],
                'min': self.fields[1][2],
                'max': self.fields[1][3]
            },
            'description': self.description
        }

        return node_dict

@dataclass
class LanguageGraphAttackStep:
    name: str = None
    type: str = None
    asset: List[ForwardRef('LanguageGraphAsset')] = None
    ttc: dict = None
    children: dict = None
    parents: dict = None
    description: dict = None

    def to_dict(self):
        node_dict = {
            'name': self.name,
            'type': self.type,
            'asset': self.asset.name,
            'ttc': self.ttc,
            'children': {},
            'parents': {},
            'description': self.description
        }

        def associations_chain_to_dict(associations_chain):
            if associations_chain:
                match (associations_chain[0]):
                    case 'union' | 'intersection' | 'difference':
                        return {associations_chain[0]: {
                            'left': associations_chain_to_dict(
                                associations_chain[1]),
                            'right': associations_chain_to_dict(
                                associations_chain[2])
                            }
                        }

                    case 'field':
                        association = associations_chain[2]
                        return {association.name:
                            {'fieldname': associations_chain[1],
                             'next_association':
                                 associations_chain_to_dict(
                                     associations_chain[3])
                            }
                        }

                    case 'transitive':
                        return {'transitive':
                            associations_chain_to_dict(
                                associations_chain[1])
                        }

                    case 'subType':
                        return {'subType': associations_chain[1],
                            'expression': associations_chain_to_dict(
                                associations_chain[2])
                        }

                    case _:
                        logger.error('Unknown associations chain element '
                            f'{associations_chain[0]}!')
                        return None
            else:
                return None

        for child in self.children:
            node_dict['children'][child] = []
            for (_, associations_chain) in self.children[child]:
                node_dict['children'][child].append(
                    associations_chain_to_dict(associations_chain))

        for parent in self.parents:
            node_dict['parents'][parent] = []
            for (_, associations_chain) in self.parents[parent]:
                node_dict['parents'][parent].append(
                    associations_chain_to_dict(associations_chain))

        return node_dict

class LanguageGraph:
    def __init__(self):
        self.assets = []
        self.associations = []
        self.attack_steps = []

    def save_to_file(self, filename: str):
        """
        Save the language graph to a json file.

        Arguments:
        filename        - the name of the output file
        """

        logger.info(f'Saving language graph to \"{filename}\" file.')
        serialized_assets = []
        for asset in self.assets:
            serialized_assets.append(asset.to_dict())
        serialized_associations = []
        for associations in self.associations:
            serialized_associations.append(associations.to_dict())
        serialized_attack_steps = []
        for attack_step in self.attack_steps:
            serialized_attack_steps.append(attack_step.to_dict())
        logger.debug(f'Saving {len(serialized_assets)} assets, '
            f'{len(serialized_associations)} associations, and '
            f'{len(serialized_attack_steps)} attack steps to '
            f'\"{filename}\" file')
        serialized_graph = {
            'Assets': serialized_assets,
            'Associations': serialized_associations,
            'Attack Steps': serialized_attack_steps
        }
        with open(filename, 'w', encoding='utf-8') as file:
            json.dump(serialized_graph, file, indent=4)


    def process_step_expression(self,
        lang: dict,
        target_asset,
        associations_chain,
        step_expression: dict):
        """
        Recursively process an attack step expression.

        Arguments:
        lang                - A dictionary representing the MAL language
                              specification.
        target_asset        - The asset type that this step expression should
                              apply to. Initially it will contain the asset
                              type to which the attack step belongs.
        associations_chain  - A chain of nested tuples that specify the
                              associations and set operations chain from the
                              attack step to its parent attack step.
                              Note: This was done for the parent attack step
                              because it was easier to construct recursively
                              given the left-hand first expansion of the
                              current MAL language specification.
        step_expression     - A dictionary containing the step expression.

        Return:
        A tuple triplet containing the target asset, the resulting parent
        associations chain, and the name of the attack step.
        """
        logger.debug('Processing Step Expression:\n' \
            + json.dumps(step_expression, indent = 2))

        match (step_expression['type']):
            case 'attackStep':
                # The attack step expression just adds the name of the attack
                # step. All other step expressions only modify the target
                # asset and parent associations chain.
                return (target_asset,
                    associations_chain,
                    step_expression['name'])

            case 'union' | 'intersection' | 'difference':
                # The set operators are used to combine the left hand and right
                # hand targets accordingly.
                lh_target_asset, lh_associations_chain, _ = self.process_step_expression(
                    lang, target_asset, associations_chain, step_expression['lhs'])
                rh_target_asset, rh_associations_chain, _ = self.process_step_expression(
                    lang, target_asset, associations_chain, step_expression['rhs'])

                if lh_target_asset != rh_target_asset:
                    logger.error('Set operation has different target asset '
                        'types for each side of the expression: '
                        f'{lh_target_asset.name} and {rh_target_asset.name}!')
                    return (None, None, None)

                return (lh_target_asset,
                    (step_expression['type'],
                        lh_associations_chain,
                        rh_associations_chain),
                    None)


                return (new_target_assets, None)

            case 'variable':
                # Fetch the step expression associated with the variable from
                # the language specification and resolve that.
                variable_step_expr = specification.\
                    get_variable_for_class_by_name(lang,
                        target_asset.name, step_expression['name'])
                if variable_step_expr:
                    return self.process_step_expression(
                        lang,
                        target_asset,
                        associations_chain,
                        variable_step_expr)

                else:
                    logger.error('Failed to find variable '
                        f'{step_expression["name"]} for {target_asset.name}')
                    return (None, None, None)

            case 'field':
                # Change the target asset from the current one to the associated
                # asset given the specified field name and add the parent
                # fieldname and association to the parent associations chain.
                fieldname = step_expression['name']
                if not target_asset:
                    logger.error(f'Missing target asset for field \"{fieldname}\"!')
                    return (None, None, None)

                for association in target_asset.associations:
                    for side in range (0, 2):
                        if association.fields[side][1] == fieldname and \
                            target_asset.subclasses(
                                association.fields[1 - side][0]):
                            return (association.fields[side][0],
                                ('field',
                                    association.fields[1 - side][1],
                                    association,
                                    associations_chain
                                ),
                                None)
                logger.error(f'Failed to find field \"{fieldname}\" on '
                    f'asset \"{target_asset.name}\"!')
                return (None, None, None)

            case 'transitive':
                # Create a transitive tuple entry that applies to the next
                # component of the step expression.
                result_target_asset, \
                result_associations_chain, \
                attack_step = \
                    self.process_step_expression(lang,
                        target_asset,
                        associations_chain,
                        step_expression['stepExpression'])
                return (result_target_asset,
                    ('transitive',
                        result_associations_chain
                    ),
                    attack_step)

            case 'subType':
                # Create a subType tuple entry that applies to the next
                # component of the step expression and changes the target
                # asset to the subasset.
                subtype_name = step_expression['subType']
                result_target_asset, \
                result_associations_chain, \
                attack_step = \
                    self.process_step_expression(lang,
                        target_asset,
                        associations_chain,
                        step_expression['stepExpression'])

                subtype_asset = next((asset for asset in self.assets \
                    if asset.name == subtype_name), None)
                if not subtype_asset:
                    logger.error('Failed to find subtype attack step '
                        f'\"{subtype_name}\"')
                if not subtype_asset.subclasses(result_target_asset):
                    logger.error(f'Found subtype \"{subtype_name}\" which '
                        f'does not extend \"{result_target_asset.name}\". '
                        'Therefore the subtype cannot be resolved.')
                    return (None, None, None)
                return (subtype_asset,
                    ('subType',
                        subtype_name,
                        result_associations_chain
                    ),
                    attack_step)

            case 'collect':
                # Apply the right hand step expression to left hand step
                # expression target asset and parent associations chain.
                (lh_target_asset, lh_associations_chain, _) = \
                    self.process_step_expression(lang,
                        target_asset,
                        associations_chain,
                        step_expression['lhs'])
                (rh_target_asset,
                    rh_associations_chain,
                    rh_attack_step_name) = \
                    self.process_step_expression(lang,
                        lh_target_asset,
                        lh_associations_chain,
                        step_expression['rhs'])
                return (rh_target_asset, rh_associations_chain,
                    rh_attack_step_name)

            case _:
                logger.error('Unknown attack step type: '
                    f'{step_expression["type"]}')
                return (None, None, None)

    def reverse_associations_chain(self, associations_chain, reverse_chain):
        """
        Recursively reverse the associations chain. From parent to child or
        vice versa.

        Arguments:
        associations_chain  - A chain of nested tuples that specify the
                              associations and set operations chain from an
                              attack step to its connected attack step.
        reverse_chain       - A chain of nested tuples that represents the
                              current reversed associations chain.

        Return:
        The resulting reversed associations chain.
        """
        if not associations_chain:
            return reverse_chain
        else:
            match (associations_chain[0]):
                case 'union' | 'intersection' | 'difference':
                    left_reverse_chain = \
                        self.reverse_associations_chain(associations_chain[1],
                        reverse_chain)
                    right_reverse_chain = \
                        self.reverse_associations_chain(associations_chain[2],
                        reverse_chain)
                    return (associations_chain[0], left_reverse_chain,
                        right_reverse_chain)

                case 'transitive':
                    return ('transitive',  self.reverse_associations_chain(
                        associations_chain[1], reverse_chain))

                case 'field':
                    association = associations_chain[2]
                    for side in range (0, 2):
                        if association.fields[side][1] == associations_chain[1]:
                            opposite_fieldname = association.fields[1 - side][1]
                            break
                    reverse_chain = ('field', opposite_fieldname, association, reverse_chain)
                    return self.reverse_associations_chain(associations_chain[3],
                        reverse_chain)

                case 'subType':
                    return ('subType', associations_chain[1],
                        self.reverse_associations_chain(
                        associations_chain[2], reverse_chain))

                case _:
                    logger.error('Unknown associations chain element '
                        f'{associations_chain[0]}!')
                    return None


    def generate_graph(self, lang: dict):
        """
        Generate language graph starting from a MAL language specification

        Arguments:
        lang            - a dictionary representing the MAL language specification
        """
        # Generate all of the asset nodes of the language graph.
        for asset in lang['assets']:
            logger.debug(f'Create asset language graph nodes for asset '
                f'{asset["name"]}')
            asset_node = LanguageGraphAsset(
                name = asset['name'],
                associations = [],
                attack_steps = [],
                description = asset['meta'],
                super_assets = [],
                sub_assets = []
            )
            self.assets.append(asset_node)

        # Link assets based on inheritance
        for asset_info in lang['assets']:
            asset = next((asset for asset in self.assets \
                if asset.name == asset_info['name']), None)
            if not asset:
                logger.error('Failed to find asset '
                    f'\"{asset_info["name"]}\"!')
                return 1
            if asset_info['superAsset']:
                super_asset = next((asset for asset in self.assets \
                    if asset.name == asset_info['superAsset']), None)
                if not super_asset:
                    logger.error('Failed to find super asset '
                        f'\"{asset_info["superAsset"]}\" '
                        f'for asset \"{asset_info["name"]}\"!')
                    return 1
                super_asset.sub_assets.append(asset)
                asset.super_assets.append(super_asset)

        # Generate all of the association nodes of the language graph.
        for asset in self.assets:
            logger.debug(f'Create association language graph nodes for asset '
                f'{asset.name}')
            associations_nodes = []
            associations = specification.get_associations_for_class(lang,
                asset.name)
            for association in associations:
                left_asset = next((asset for asset in self.assets \
                    if asset.name == association['leftAsset']), None)
                if not left_asset:
                    logger.error('Failed to find left hand asset '
                        f'\"{association["leftAsset"]}\" for '
                        f'association \"{association["name"]}\"!')
                    return 1
                right_asset = next((asset for asset in self.assets \
                    if asset.name == association['rightAsset']), None)
                if not right_asset:
                    logger.error('Failed to find right hand asset '
                        f'\"{association["rightAsset"]}\" for '
                        f'association \"{association["name"]}\"!')
                    return 1

                assoc_node = next((assoc for assoc in self.associations \
                    if assoc.name == association['name'] and
                        assoc.fields[0][0] == left_asset and
                        assoc.fields[1][0] == right_asset),
                        None)
                if assoc_node:
                    # The association was already created, skip it
                    continue

                assoc_node = LanguageGraphAssociation(
                    name = association['name'],
                    fields = (
                        (left_asset,
                            association['leftField'],
                            association['leftMultiplicity']['min'],
                            association['leftMultiplicity']['max']
                        ),
                        (right_asset,
                            association['rightField'],
                            association['rightMultiplicity']['min'],
                            association['rightMultiplicity']['max']
                        )
                    ),
                    description = association['meta']
                )

                # Add the association to the left and right asset and all of
                # the assets that inherit them
                associated_assets = [left_asset, right_asset]
                while associated_assets != []:
                    asset = associated_assets.pop()
                    associated_assets.extend(asset.sub_assets)
                    if assoc_node not in asset.associations:
                        asset.associations.append(assoc_node)

                self.associations.append(assoc_node)

        # Generate all of the attack step nodes of the language graph.
        for asset in self.assets:
            logger.debug(f'Create attack steps language graph nodes for asset '
                f'{asset.name}.')
            attack_step_nodes = []
            attack_steps = specification.get_attacks_for_class(lang,
                asset.name)
            for attack_step_name, attack_step_attribs in attack_steps.items():
                logger.debug(f'Create attack step language graph nodes for '
                    f'{attack_step_name}.')

                attack_step_node = LanguageGraphAttackStep(
                    name = asset.name + ':' + attack_step_name,
                    type = attack_step_attribs['type'],
                    asset = asset,
                    ttc = attack_step_attribs['ttc'],
                    children = {},
                    parents = {},
                    description = attack_step_attribs['meta']
                )
                attack_step_node.attributes = attack_step_attribs
                asset.attack_steps.append(attack_step_node)
                self.attack_steps.append(attack_step_node)

        # Then, link all of the attack step nodes according to their associations.
        for attack_step in self.attack_steps:
            logger.debug('Determining children for attack step '\
                f'{attack_step.name}.')
            step_expressions = \
                attack_step.attributes['reaches']['stepExpressions'] if \
                    attack_step.attributes['reaches'] else []

            for step_expression in step_expressions:
                # Resolve each of the attack step expressions listed for this
                # attack step to determine children.
                (target_asset, associations_chain, attack_step_name) = \
                    self.process_step_expression(lang,
                        attack_step.asset,
                        None,
                        step_expression)
                if not target_asset:
                    logger.error('Failed to find target asset ' \
                    f'to link with for step expression:\n' +
                    json.dumps(step_expression, indent = 2))
                    print('Failed to find target asset ' \
                    f'to link with for step expression:\n' +
                    json.dumps(step_expression, indent = 2))
                    return 1

                attack_step_fullname = target_asset.name + ':' + attack_step_name
                target_attack_step = next((attack_step \
                    for attack_step in target_asset.attack_steps \
                        if attack_step.name == attack_step_fullname), None)

                if not target_attack_step:
                    logger.error('Failed to find target attack step on '
                        f'{target_asset.name} to link with for step '
                        'expression:\n' +
                        json.dumps(step_expression, indent = 2))
                    print('Failed to find target attack step on '
                        f'{target_asset.name} to link with for step '
                        'expression:\n' +
                        json.dumps(step_expression, indent = 2))
                    return 1

                # It is easier to create the parent associations chain due to
                # the left-hand first progression.
                if attack_step.name in target_attack_step.parents:
                    target_attack_step.parents[attack_step.name].append(
                        (target_attack_step, associations_chain))
                else:
                    target_attack_step.parents[attack_step.name] = \
                        [(target_attack_step, associations_chain)]
                # Reverse the parent associations chain to get the child
                # associations chain.
                if target_attack_step.name in attack_step.children:
                    attack_step.children[target_attack_step.name].append(
                        (attack_step,
                        self.reverse_associations_chain(associations_chain,
                            None)))
                else:
                    attack_step.children[target_attack_step.name] = \
                        [(attack_step,
                        self.reverse_associations_chain(associations_chain,
                            None))]

        return 0
